# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Module to use to extract archives. No business logic."""

import concurrent.futures
import contextlib
import gzip
import io
import os
import shutil
import tarfile
import uuid
import zipfile

import promise

from .. import utils
from ..utils.file_utils import INCOMPLETE_SUFFIX
from . import resource as resource_lib

import bz2


@utils.memoize()
def get_extractor(*args, **kwargs):
	return _Extractor(*args, **kwargs)


class ExtractError(Exception):
	"""There was an error while extracting the archive."""


class UnsafeArchiveError(Exception):
	"""The archive is unsafe to unpack, e.g. absolute path."""


class _Extractor(object):
	"""Singleton (use `get_extractor()` module fct) to extract archives."""

	def __init__(self, max_workers=12):
		self._executor = concurrent.futures.ThreadPoolExecutor(
				max_workers=max_workers)
		self._pbar_path = None

	@contextlib.contextmanager
	def tqdm(self):
		"""Add a progression bar for the current extraction."""
		with utils.async_tqdm(
				total=0, desc='Extraction completed...', unit=' file') as pbar_path:
			self._pbar_path = pbar_path
			yield

	def extract(self, path, extract_method, to_path):
		"""Returns `promise.Promise` => to_path."""
		self._pbar_path.update_total(1)
		if extract_method not in _EXTRACT_METHODS:
			raise ValueError('Unknown extraction method "%s".' % extract_method)
		future = self._executor.submit(self._sync_extract,
																	 path, extract_method, to_path)
		return promise.Promise.resolve(future)

	def _sync_extract(self, from_path, method, to_path):
		"""Returns `to_path` once resource has been extracted there."""
		to_path_tmp = '%s%s_%s' % (to_path, INCOMPLETE_SUFFIX, uuid.uuid4().hex)
		path = None
		try:
			for path, handle in iter_archive(from_path, method):
				path = str(path)
				_copy(handle, path and os.path.join(to_path_tmp, path) or to_path_tmp)
		except BaseException as err:
			msg = 'Error while extracting %s to %s (file: %s) : %s' % (
					from_path, to_path, path, err)
			raise ExtractError(msg)
		# `os.rename(overwrite=True)` doesn't work for non empty
		# directories, so delete destination first, if it already exists.
		if os.path.exists(to_path):
			shutil.rmtree(to_path)
		os.rename(to_path_tmp, to_path)
		self._pbar_path.update(1)
		return to_path


def _copy(src_file, dest_path):
	"""Copy data read from src file obj to new file in dest_path."""
	os.makedirs(os.path.dirname(dest_path))
	with open(dest_path, 'wb') as dest_file:
		while True:
			data = src_file.read(io.DEFAULT_BUFFER_SIZE)
			if not data:
				break
			dest_file.write(data)


def _normpath(path):
	path = os.path.normpath(path)
	if (path.startswith('.')
			or os.path.isabs(path)
			or path.endswith('~')
			or os.path.basename(path).startswith('.')):
		return None
	return path


@contextlib.contextmanager
def _open_or_pass(path_or_fobj):
	if isinstance(path_or_fobj, str):
		with open(path_or_fobj, 'rb') as f_obj:
			yield f_obj
	else:
		yield path_or_fobj


def iter_tar(arch_f, stream=False):
	"""Iter over tar archive, yielding (path, object-like) tuples.

	Args:
		arch_f: File object of the archive to iterate.
		stream: If True, open the archive in stream mode which allows for faster
			processing and less temporary disk consumption, but random access to the
			file is not allowed.

	Yields:
		(filepath, extracted_fobj) for each file in the archive.
	"""
	read_type = 'r' + ('|' if stream else ':') + '*'

	with _open_or_pass(arch_f) as fobj:
		tar = tarfile.open(mode=read_type, fileobj=fobj)
		for member in tar:
			extract_file = tar.extractfile(member)
			if extract_file:  # File with data (not directory):
				path = _normpath(member.path)
				if not path:
					continue
				yield [path, extract_file]


def iter_tar_stream(arch_f):
	return iter_tar(arch_f, stream=True)


def iter_gzip(arch_f):
	with _open_or_pass(arch_f) as fobj:
		gzip_ = gzip.GzipFile(fileobj=fobj)
		yield ('', gzip_)  # No inner file.


def iter_bzip2(arch_f):
	with _open_or_pass(arch_f) as fobj:
		bz2_ = bz2.BZ2File(filename=fobj)
		yield ('', bz2_)  # No inner file.


def iter_zip(arch_f):
	with _open_or_pass(arch_f) as fobj:
		z = zipfile.ZipFile(fobj)
		for member in z.infolist():
			extract_file = z.open(member)
			if extract_file:  # File with data (not directory):
				path = _normpath(member.filename)
				if not path:
					continue
				yield [path, extract_file]


_EXTRACT_METHODS = {
		resource_lib.ExtractMethod.BZIP2: iter_bzip2,
		resource_lib.ExtractMethod.GZIP: iter_gzip,
		resource_lib.ExtractMethod.TAR: iter_tar,
		resource_lib.ExtractMethod.TAR_GZ: iter_tar,
		resource_lib.ExtractMethod.TAR_GZ_STREAM: iter_tar_stream,
		resource_lib.ExtractMethod.TAR_STREAM: iter_tar_stream,
		resource_lib.ExtractMethod.ZIP: iter_zip,
}


def iter_archive(path, method):
	"""Yields (path_in_archive, f_obj) for archive at path using `nlp.download.ExtractMethod`."""  # pylint: disable=line-too-long
	return _EXTRACT_METHODS[method](path)



# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Data downloads using the Kaggle CLI."""

import collections
import os
import subprocess as sp

import logging

from .. import utils

logger = logging.getLogger(__name__)

_ERR_MSG = """\
To download Kaggle data, follow the instructions to install the \
kaggle API and get API credentials:
https://github.com/Kaggle/kaggle-api#installation

Additionally, you must join the competition through the Kaggle competition's \
website:
https://www.kaggle.com/c/%s
"""

_NOT_FOUND_ERR_MSG = """\
Competition %s not found. Please ensure you have spelled the competition name \
correctly.
"""

KaggleType = collections.namedtuple(
        "KaggleType",
        ["prefix", "download_cmd", "dl_flag", "extra_flag"])

_KAGGLE_TYPES = {
        "dataset": KaggleType(
                prefix="dataset",
                download_cmd="datasets",
                dl_flag="-d",
                extra_flag="--unzip"),
        "competition": KaggleType(
                prefix="competition",
                download_cmd="competitions",
                dl_flag="-c",
                extra_flag="")
}


def _get_kaggle_type(competition_name):
    if "/" in competition_name:
        return _KAGGLE_TYPES["dataset"]
    else:
        return _KAGGLE_TYPES["competition"]


class KaggleFile(object):
    """Represents a Kaggle competition file."""
    _URL_PREFIX = "kaggle://"

    def __init__(self, competition_name, filename):
        self._competition_name = competition_name
        self._filename = filename
        self.type = _get_kaggle_type(competition_name)

    @property
    def competition(self):
        return self._competition_name

    @property
    def filename(self):
        return self._filename

    @classmethod
    def from_url(cls, url):
        if not KaggleFile.is_kaggle_url(url):
            raise TypeError("Not a valid kaggle URL")
        download_type, competition_name, filename = (
                url[len(cls._URL_PREFIX):].split("/", 2))
        if download_type == "dataset":
            dataset_name, filename = filename.split("/", 1)
            competition_name = "%s/%s" % (competition_name, dataset_name)
        return cls(competition_name, filename)

    @staticmethod
    def is_kaggle_url(url):
        return url.startswith(KaggleFile._URL_PREFIX)

    def to_url(self):
        return "%s%s/%s/%s" % (self._URL_PREFIX,
                                                     self.type.prefix,
                                                     self._competition_name,
                                                     self._filename)


class KaggleCompetitionDownloader(object):
    """Downloader for a Kaggle competition.

    Usage:
    You can download with dataset or competition name like `zillow/zecon`
    or `titanic`.

    ```
    downloader = KaggleCompetitionDownloader(competition_name)
    for fname in downloader.competition_files:
        downloader.download_file(fname, make_file_output_path(fname))
    ```
    """

    def __init__(self, competition_name):
        self._competition_name = competition_name
        self._kaggle_type = _get_kaggle_type(self._competition_name)

    @utils.memoized_property
    def competition_files(self):
        """List of competition files."""
        command = [
                "kaggle",
                self._kaggle_type.download_cmd,
                "files",
                "-v",
                self._competition_name,
        ]
        output = _run_kaggle_command(command, self._competition_name)
        return sorted([
                line.split(",")[0] for line in output.split("\n")[1:] if line
        ])

    @utils.memoized_property
    def competition_urls(self):
        """Returns 'kaggle://' urls."""
        return [
                KaggleFile(self._competition_name, fname).to_url()
                for fname in self.competition_files  # pylint: disable=not-an-iterable
        ]

    def download_file(self, fname, output_dir):
        """Downloads competition file to output_dir."""
        if fname not in self.competition_files:  # pylint: disable=unsupported-membership-test
            raise ValueError("%s is not one of the competition's "
                                             "files: %s" % (fname, self.competition_files))
        command = [
                "kaggle",
                self._kaggle_type.download_cmd,
                "download",
                "--file",
                fname,
                "--path",
                output_dir,
                self._kaggle_type.dl_flag,
                self._competition_name
        ]
        if self._kaggle_type.extra_flag:
            command.append(self._kaggle_type.extra_flag)
        _run_kaggle_command(command, self._competition_name)
        return os.path.join(output_dir, fname)


def _run_kaggle_command(command_args, competition_name):
    """Run kaggle command with subprocess."""
    try:
        output = sp.check_output(command_args)
        return str(output)
    except sp.CalledProcessError as err:
        output = err.output
        _log_command_output(output, error=True)
        if output.startswith(b"404"):
            logger.error(_NOT_FOUND_ERR_MSG, competition_name)
            raise
        logger.error(_ERR_MSG, competition_name)
        raise


def _log_command_output(output, error=False):
    log = logger.error if error else logger.info
    log("kaggle command output:\n%s", str(output))



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Version utils."""

import enum
import re



_VERSION_TMPL = (
        r"^(?P<major>{v})"
        r"\.(?P<minor>{v})"
        r"\.(?P<patch>{v})$")
_VERSION_WILDCARD_REG = re.compile(_VERSION_TMPL.format(v=r"\d+|\*"))
_VERSION_RESOLVED_REG = re.compile(_VERSION_TMPL.format(v=r"\d+"))


class Experiment(enum.Enum):
    """Experiments which can be enabled/disabled on a per version basis.

    Experiments are designed to gradually apply changes to datasets while
    maintaining backward compatibility with previous versions. All experiments
    should eventually be deleted, once used by all versions of all datasets.

    Eg:
    class Experiment(enum.Enum):
        EXP_A = enum.auto()  # Short description of experiment.

    class MyBuilder(...):
        VERSION = nlp.Version('1.2.3', experiments={
                nlp.Experiment.EXP_A: True,
                })
    """
    # A Dummy experiment, which should NOT be used, except for testing.
    DUMMY = 1

    # New Shuffling, sharding and slicing mechanism.
    S3 = 2

    # Experiment to update the dataset metadata files without making the
    # new dataset version the default.
    METADATA = 3


class Version(object):
    """Dataset version MAJOR.MINOR.PATCH."""

    _DEFAULT_EXPERIMENTS = {
            Experiment.DUMMY: False,
            Experiment.S3: True,
            Experiment.METADATA: False,
    }

    def __init__(self, version_str, description=None, experiments=None,
                             nlp_version_to_prepare=None):
        """Version init.

        Args:
            version_str: string. Eg: "1.2.3".
            description: string, a description of what is new in this version.
            experiments: dict of experiments. See Experiment.
            nlp_version_to_prepare: string, defaults to None. If set, indicates that
                current version cannot be used to `download_and_prepare` the
                dataset, but that at version {nlp_version_to_prepare} should be
                used instead.
        """
        if description is not None and not isinstance(description, str):
            raise TypeError(
                    "Description should be a string. Got {}".format(description))
        self.description = description
        self._experiments = self._DEFAULT_EXPERIMENTS.copy()
        self.nlp_version_to_prepare = nlp_version_to_prepare
        if experiments:
            self._experiments.update(experiments)
        self.major, self.minor, self.patch = _str_to_version(version_str)

    def implements(self, experiment):
        """Returns True if version implements given experiment."""
        return self._experiments[experiment]

    def __str__(self):
        return "{}.{}.{}".format(*self.tuple)

    @property
    def tuple(self):
        return self.major, self.minor, self.patch

    def _validate_operand(self, other):
        if isinstance(other, str):
            return Version(other)
        elif isinstance(other, Version):
            return other
        raise AssertionError("{} (type {}) cannot be compared to version.".format(
                other, type(other)))

    def __eq__(self, other):
        other = self._validate_operand(other)
        return self.tuple == other.tuple

    def __ne__(self, other):
        other = self._validate_operand(other)
        return self.tuple != other.tuple

    def __lt__(self, other):
        other = self._validate_operand(other)
        return self.tuple < other.tuple

    def __le__(self, other):
        other = self._validate_operand(other)
        return self.tuple <= other.tuple

    def __gt__(self, other):
        other = self._validate_operand(other)
        return self.tuple > other.tuple

    def __ge__(self, other):
        other = self._validate_operand(other)
        return self.tuple >= other.tuple

    def match(self, other_version):
        """Returns True if other_version matches.

        Args:
            other_version: string, of the form "x[.y[.x]]" where {x,y,z} can be a
                number or a wildcard.
        """
        major, minor, patch = _str_to_version(other_version, allow_wildcard=True)
        return (major in [self.major, "*"] and minor in [self.minor, "*"]
                        and patch in [self.patch, "*"])


def _str_to_version(version_str, allow_wildcard=False):
    """Return the tuple (major, minor, patch) version extracted from the str."""
    reg = _VERSION_WILDCARD_REG if allow_wildcard else _VERSION_RESOLVED_REG
    res = reg.match(version_str)
    if not res:
        msg = "Invalid version '{}'. Format should be x.y.z".format(version_str)
        if allow_wildcard:
            msg += " with {x,y,z} being digits or wildcard."
        else:
            msg += " with {x,y,z} being digits."
        raise ValueError(msg)
    return tuple(
            v if v == "*" else int(v)
            for v in [res.group("major"), res.group("minor"), res.group("patch")])



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Wrapper around FeatureDict to allow better control over decoding.
"""

from .feature import FeatureConnector


class TopLevelFeature(FeatureConnector):
    """Top-level `FeatureConnector` to manage decoding.

    Note that `FeatureConnector` which are declared as `TopLevelFeature` can be
    nested. However, only the top-level feature can be decoded.

    `TopLevelFeature` allows better control over the decoding, and
    eventually better support for augmentations.
    """

    def __init__(self, *args, **kwargs):
        """Constructor."""
        self.__is_top_level = False
        super(TopLevelFeature, self).__init__(*args, **kwargs)

    # # AutoGraph doesn't support mangled names (__is_top_level), so we explicitly
    # # disable it in methods that use them, to avoid the warning.
    # # TODO(mdan): Remove decorator once AutoGraph supports mangled names.
    # @tf.autograph.experimental.do_not_convert()
    def _set_top_level(self):
        """Indicates that the feature is top level.

        Internal function called by `DatasetInfo`.
        """
        self.__is_top_level = True

    # # AutoGraph doesn't support mangled names (__is_top_level), so we explicitly
    # # disable it in methods that use them, to avoid the warning.
    # # TODO(mdan): Remove decorator once AutoGraph supports mangled names.
    # @tf.autograph.experimental.do_not_convert()
    def decode_example(self, serialized_example, decoders=None):
        # pylint: disable=line-too-long
        """Decode the serialize examples.

        Args:
            serialized_example: Nested `dict` of `tf.Tensor`
            decoders: Nested dict of `Decoder` objects which allow to customize the
                decoding. The structure should match the feature structure, but only
                customized feature keys need to be present. See
                [the guide](https://github.com/tensorflow/datasets/tree/master/docs/decode.md)
                for more info.

        Returns:
            example: Nested `dict` containing the decoded nested examples.
        """
        # pylint: enable=line-too-long
        if not self.__is_top_level:
            raise AssertionError(
                    'Feature {} can only be decoded when defined as top-level '
                    'feature, through info.features.decode_example()'.format(
                            type(self).__name__))

        # Step 1: Flatten the nested dict => []
        flat_example = self._flatten(serialized_example)
        flat_features = self._flatten(self)
        flat_serialized_info = self._flatten(self.get_type())
        flat_decoders = self._flatten(decoders)

        # Step 2: Apply the decoding
        flatten_decoded = []
        for (
                feature,
                example,
                serialized_info,
                decoder,
        ) in zip(
                flat_features,
                flat_example,
                flat_serialized_info,
                flat_decoders):
            flatten_decoded.append(_decode_feature(
                    feature=feature,
                    example=example,
                    serialized_info=serialized_info,
                    decoder=decoder,
            ))

        # Step 3: Restore nesting [] => {}
        nested_decoded = self._nest(flatten_decoded)
        return nested_decoded


def _decode_feature(feature, example, serialized_info, decoder):
    """Decode a single feature."""
    # Eventually overwrite the default decoding
    if decoder is not None:
        decoder.setup(feature=feature)
    else:
        decoder = feature

    sequence_rank = _get_sequence_rank(serialized_info)
    if sequence_rank == 0:
        return decoder.decode_example(example)
    elif sequence_rank == 1:
        # Return a batch of examples from a sequence
        return decoder.decode_batch_example(example)
    elif sequence_rank > 1:
        # Use ragged tensor if the sequance rank is greater than one
        return decoder.decode_ragged_example(example)


def _get_sequence_rank(serialized_info):
    """Return the number of sequence dimensions of the feature."""
    if isinstance(serialized_info, dict):
        all_sequence_rank = [s.sequence_rank for s in serialized_info.values()]
    else:
        all_sequence_rank = [serialized_info.sequence_rank]

    sequence_ranks = set(all_sequence_rank)
    if len(sequence_ranks) != 1:
        raise NotImplementedError(
                'Decoding do not support mixing sequence and context features within a '
                'single FeatureConnector. Received inputs of different sequence_rank: '
                '{}'.format(sequence_ranks)
        )
    return next(iter(sequence_ranks))



import os
import re
import shutil
from argparse import ArgumentParser, Namespace
from logging import getLogger

from nlp.commands import BaseTransformersCLICommand


def convert_command_factory(args: Namespace):
    """
    Factory function used to convert a model TF 1.0 checkpoint in a PyTorch checkpoint.
    :return: ServeCommand
    """
    return ConvertCommand(
        args.tfds_directory, args.nlp_directory
    )


class ConvertCommand(BaseTransformersCLICommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        """
        Register this command to argparse so it's available for the transformer-cli
        :param parser: Root parser to register command-specific arguments
        :return:
        """
        train_parser = parser.add_parser(
            "convert",
            help="CLI tool to convert a (nlp) TensorFlow-Dataset in a HuggingFace-NLP dataset.",
        )
        train_parser.add_argument(
            "--tfds_directory", type=str, required=True, help="Path to the TensorFlow Datasets folder."
        )
        train_parser.add_argument(
            "--nlp_directory", type=str, required=True, help="Path to the HuggingFace NLP folder."
        )
        train_parser.set_defaults(func=convert_command_factory)

    def __init__(
        self,
        tfds_directory: str,
        nlp_directory: str,
        *args
    ):
        self._logger = getLogger("nlp-cli/converting")

        self._tfds_directory = tfds_directory
        self._nlp_directory = nlp_directory

    def run(self):
        abs_tfds_path = os.path.abspath(self._tfds_directory)
        abs_nlp_path = os.path.abspath(self._nlp_directory)
        self._logger.info("Converting datasets from %s to %s", abs_tfds_path, abs_nlp_path)

        utils_files = []
        imports_to_builder_map = {}
        for f_name in os.listdir(abs_tfds_path):
            self._logger.info("Looking at file %s", f_name)
            input_file = os.path.join(abs_tfds_path, f_name)
            output_file = os.path.join(abs_nlp_path, f_name)

            if not os.path.isfile(input_file) or '__init__' in f_name or '_test' in f_name or not '.py' in f_name:
                self._logger.info("Skipping file")
                continue

            with open(input_file, "r") as f:
                lines = f.readlines()

            out_lines = []
            is_builder = False
            tfds_imports = []
            for line in lines:
                out_line = line

                # Convert imports
                if 'import tensorflow.compat.v2 as tf' in out_line:
                    continue
                elif '@tfds.core' in out_line:
                    continue
                elif 'import tensorflow_datasets.public_api as tfds' in out_line:
                    out_line = 'import nlp\n'
                elif 'from absl import logging' in out_line:
                    out_line = 'import logging\n'
                else:
                    out_line = out_line.replace('tfds.core', 'nlp')
                    out_line = out_line.replace('tf.io.gfile.GFile', 'open')
                    out_line = out_line.replace('tf.', 'nlp.')
                    out_line = out_line.replace('tfds.features.Text()', 'nlp.string')
                    out_line = out_line.replace('The TensorFlow Datasets Authors',
                                                'The TensorFlow Datasets Authors and the HuggingFace NLP Authors')
                    out_line = out_line.replace('tfds.', 'nlp.')

                # Take care of saving utilities (to later move them together with main script)
                if 'tensorflow_datasets' in out_line:
                    match = re.match(r"from\stensorflow_datasets.*import\s([^\.\r\n]+)", out_line)
                    tfds_imports.extend(imp.strip() for imp in match.group(1).split(','))
                    out_line = 'from . import ' + match.group(1)

                # Check we have not forget anything
                assert not 'tf.' in out_line and not 'tfds.' in out_line and not 'tensorflow_datasets' in out_line, f"Error converting {out_line.strip()}"

                if "GeneratorBasedBuilder" in out_line or "BeamBasedBuilder" in out_line:
                    is_builder = True
                out_lines.append(out_line)

            if is_builder:
                # We create a new directory for each dataset
                dir_name = f_name.replace('.py', '')
                output_dir = os.path.join(abs_nlp_path, dir_name)
                output_file = os.path.join(output_dir, f_name)
                os.makedirs(output_dir, exist_ok=True)
                self._logger.info("Adding directory %s", output_dir)
                imports_to_builder_map.update(dict((imp, output_dir) for imp in tfds_imports))
            else:
                # Utilities will be moved at the end
                utils_files.append(output_file)

            with open(output_file, "w") as f:
                f.writelines(out_lines)
            self._logger.info("Converted in %s", output_file)

        for utils_file in utils_files:
            try:
                f_name = os.path.basename(utils_file)
                dest_folder = imports_to_builder_map[f_name.replace('.py', '')]
                self._logger.info("Moving %s to %s", utils_file, dest_folder)
                shutil.copy(utils_file, dest_folder)
            except KeyError:
                self._logger.error(f"Cannot find destination folder for {utils_file}. Please copy manually.")



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Text feature.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import pyarrow as pa

from .feature import Tensor
# from .text import SubwordTextEncoder, TextEncoderConfig


class Text(Tensor):
    """`FeatureConnector` for text, encoding to integers with a `TextEncoder`."""

    def __init__(self, encoder=None, encoder_config=None):
        """Constructs a Text FeatureConnector.

        Args:
            encoder: `nlp.features.text.TextEncoder`, an encoder that can convert
                text to integers. If None, the text will be utf-8 byte-encoded.
            encoder_config: `nlp.features.text.TextEncoderConfig`, needed if
                restoring from a file with `load_metadata`.
        """
        if encoder and encoder_config:
            raise ValueError("If encoder is provided, encoder_config must be None.")
        # if encoder:
        #     encoder_config = TextEncoderConfig(
        #             encoder_cls=type(encoder),
        #             vocab_size=encoder.vocab_size)
        # elif encoder_config:
        #     encoder = encoder_config.encoder

        self._encoder = encoder
        self._encoder_config = encoder_config

        has_encoder = bool(encoder or self._encoder_cls)
        super(Text, self).__init__(
                shape=(None,) if has_encoder else (),
                dtype=pa.int64 if has_encoder else pa.string,
        )

    @property
    def encoder(self):
        return self._encoder

    @encoder.setter
    def encoder(self, new_encoder):
        if self.encoder:
            raise ValueError("Cannot override encoder")
        self._encoder = new_encoder
        encoder_cls = self._encoder_cls or type(None)
        if not isinstance(new_encoder, encoder_cls):
            raise ValueError(
                    "Changing type of encoder. Got %s but must be %s" %
                    (type(new_encoder).__name__,
                     self._encoder_cls.__name__))

    def maybe_set_encoder(self, new_encoder):
        """Set encoder, but no-op if encoder is already set."""
        if self.encoder:
            return
        self.encoder = new_encoder

    @property
    def vocab_size(self):
        return self.encoder and self.encoder.vocab_size

    def str2ints(self, str_value):
        """Conversion string => encoded list[int]."""
        if not self._encoder:
            raise ValueError(
                    "Text.str2ints is not available because encoder hasn't been defined.")
        return self._encoder.encode(str_value)

    def ints2str(self, int_values):
        """Conversion list[int] => decoded string."""
        if not self._encoder:
            raise ValueError(
                    "Text.ints2str is not available because encoder hasn't been defined.")
        return self._encoder.decode(int_values)

    def encode_example(self, example_data):
        if self.encoder:
            example_data = self.encoder.encode(example_data)
        else:
            example_data = example_data
        return super(Text, self).encode_example(example_data)

    def save_metadata(self, data_dir, feature_name):
        fname_prefix = os.path.join(data_dir, "%s.text" % feature_name)
        if not self.encoder:
            return
        self.encoder.save_to_file(fname_prefix)

    def load_metadata(self, data_dir, feature_name):
        fname_prefix = os.path.join(data_dir, "%s.text" % feature_name)
        encoder_cls = self._encoder_cls
        if encoder_cls:
            self._encoder = encoder_cls.load_from_file(fname_prefix)
            return

        # Error checking: ensure there are no metadata files
        feature_files = [
                f for f in os.listdir(data_dir) if f.startswith(fname_prefix)
        ]
        if feature_files:
            raise ValueError(
                    "Text feature files found for feature %s but encoder_cls=None. "
                    "Make sure to set encoder_cls in the TextEncoderConfig. "
                    "Files: %s" % (feature_name, feature_files))

    def maybe_build_from_corpus(self, corpus_generator, **kwargs):
        """Call SubwordTextEncoder.build_from_corpus is encoder_cls is such.

        If `self.encoder` is `None` and `self._encoder_cls` is of type
        `SubwordTextEncoder`, the method instantiates `self.encoder` as returned
        by `SubwordTextEncoder.build_from_corpus()`.

        Args:
            corpus_generator: generator yielding `str`, from which
                subwords will be constructed.
            **kwargs: kwargs forwarded to `SubwordTextEncoder.build_from_corpus()`
        """
        pass
        # if self._encoder_cls is not SubwordTextEncoder:
        #     return
        # if self.encoder:
        #     return

        # vocab_size = self._encoder_config.vocab_size
        # self.encoder = SubwordTextEncoder.build_from_corpus(
        #         corpus_generator=corpus_generator,
        #         target_vocab_size=vocab_size,
        #         **kwargs)

    @property
    def _encoder_cls(self):
        return self._encoder_config and self._encoder_config.encoder_cls

    def _additional_repr_info(self):
        if self.encoder is None:
            return {}
        return {"encoder": repr(self.encoder)}



# Lint as: python3
"""HuggingFace/NLP is an open library of NLP datasets.

Simple check list for release from AllenNLP repo: https://github.com/allenai/allennlp/blob/master/setup.py

To create the package for pypi.

1. Change the version in __init__.py, setup.py as well as docs/source/conf.py.

2. Commit these changes with the message: "Release: VERSION"

3. Add a tag in git to mark the release: "git tag VERSION -m'Adds tag VERSION for pypi' "
   Push the tag to git: git push --tags origin master

4. Build both the sources and the wheel. Do not change anything in setup.py between
   creating the wheel and the source distribution (obviously).

   For the wheel, run: "python setup.py bdist_wheel" in the top level directory.
   (this will build a wheel for the python version you use to build it).

   For the sources, run: "python setup.py sdist"
   You should now have a /dist directory with both .whl and .tar.gz source versions.

5. Check that everything looks correct by uploading the package to the pypi test server:

   twine upload dist/* -r pypitest
   (pypi suggest using twine as other methods upload files via plaintext.)
   You may have to specify the repository url, use the following command then:
   twine upload dist/* -r pypitest --repository-url=https://test.pypi.org/legacy/

   Check that you can install it in a virtualenv by running:
   pip install -i https://testpypi.python.org/pypi transformers

6. Upload the final version to actual pypi:
   twine upload dist/* -r pypi

7. Copy the release notes from RELEASE.md to the tag in github once everything is looking hunky-dory.

8. Update the documentation commit in .circleci/deploy.sh for the accurate documentation to be displayed

9. Update README.md to redirect to correct documentation.
"""

import datetime
import itertools
import os
import sys

from setuptools import find_packages
from setuptools import setup

DOCLINES = __doc__.split('\n')

REQUIRED_PKGS = [
    'numpy',
    'pyarrow',
    'dill',
    'promise',
    'dataclasses-json',
    # for downloading datasets over HTTPS
    'requests>=2.19.0',
    # progress bars in download and scripts
    "tqdm >= 4.27",
    # dataclasses for Python versions that don't have it
    "dataclasses;python_version<'3.7'",
    # accessing files from S3 directly
    "boto3",
    # filesystem locks e.g. to prevent parallel downloads
    "filelock",
]

TESTS_REQUIRE = [
    'apache-beam',
    'jupyter',
    'pytest',
    'pytest-xdist',
]


# Extra dependencies required by specific datasets
DATASET_EXTRAS = {
    # In alphabetical order
    'aflw2k3d': ['scipy'],
    'c4': ['apache_beam', 'langdetect', 'nltk', 'tldextract'],
    'the300w_lp': ['scipy'],
    'wikipedia': ['mwparserfromhell', 'apache_beam'],
}


# Extra dataset deps are required for the tests
all_dataset_extras = list(itertools.chain.from_iterable(
    deps for ds_name, deps in DATASET_EXTRAS.items()
))


EXTRAS_REQUIRE = {
    'apache-beam': ['apache-beam'],
    'tensorflow': ['tensorflow>=1.15.0'],
    'tensorflow_gpu': ['tensorflow-gpu>=1.15.0'],
    'torch': ['torch'],
    # Tests dependencies are installed in ./oss_scripts/oss_pip_install.sh
    # and run in ./oss_scripts/oss_tests.sh
    'tests': TESTS_REQUIRE + all_dataset_extras,
}
EXTRAS_REQUIRE.update(DATASET_EXTRAS)

setup(
    name='nlp',
    version="0.0.1",
    description=DOCLINES[0],
    long_description='\n'.join(DOCLINES[2:]),
    author='HuggingFace Inc.',
    author_email='thomas@huggingface.ce',
    url='https://github.com/huggingface/nlp',
    download_url='https://github.com/huggingface/nlp/tags',
    license='Apache 2.0',
    package_dir={"": "src"},
    packages=find_packages("src"),
    package_data={
        'nlp': [
            'scripts/templates/*',
        ],
    },
    scripts=["nlp-cli"],
    install_requires=REQUIRED_PKGS,
    extras_require=EXTRAS_REQUIRE,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    keywords='nlp machine learning datasets',
)



# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Utils functions."""

import enum
import functools
import os
import threading

import urllib


class GenerateMode(enum.Enum):
    """`Enum` for how to treat pre-existing downloads and data.

    The default mode is `REUSE_DATASET_IF_EXISTS`, which will reuse both
    raw downloads and the prepared dataset if they exist.

    The generations modes:

    |                                    | Downloads | Dataset |
    | -----------------------------------|-----------|---------|
    | `REUSE_DATASET_IF_EXISTS` (default)| Reuse     | Reuse   |
    | `REUSE_CACHE_IF_EXISTS`            | Reuse     | Fresh   |
    | `FORCE_REDOWNLOAD`                 | Fresh     | Fresh   |
    """

    REUSE_DATASET_IF_EXISTS = 'reuse_dataset_if_exists'
    REUSE_CACHE_IF_EXISTS = 'reuse_cache_if_exists'
    FORCE_REDOWNLOAD = 'force_redownload'


class ComputeStatsMode(enum.Enum):
    """Mode to decide if dynamic dataset info fields should be computed or not.

    Mode can be:

    * AUTO: Compute the DatasetInfo dynamic fields only if they haven't been
        restored from GCS.
    * FORCE: Always recompute DatasetInfo dynamic  fields, even if they are
        already present
    * SKIP: Ignore the dataset dynamic field computation (whether they already
        exist or not)

    """

    AUTO = 'auto'
    FORCE = 'force'
    SKIP = 'skip'


# TODO(epot): Move some of those functions into py_utils


def build_synchronize_decorator():
    """Returns a decorator which prevents concurrent calls to functions.

    Usage:
        synchronized = build_synchronize_decorator()

        @synchronized
        def read_value():
            ...

        @synchronized
        def write_value(x):
            ...

    Returns:
        make_threadsafe (fct): The decorator which lock all functions to which it
            is applied under a same lock
    """
    lock = threading.Lock()

    def lock_decorator(fn):

        @functools.wraps(fn)
        def lock_decorated(*args, **kwargs):
            with lock:
                return fn(*args, **kwargs)

        return lock_decorated

    return lock_decorator


def get_file_name(url):
    """Returns file name of file at given url."""
    return os.path.basename(urllib.parse.urlparse(url).path) or 'unknown_name'



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""To write records into Parquet files."""
import collections
import itertools
import json
import os

from typing import Dict, List, Optional, Any

import logging

import pyarrow as pa

logger = logging.getLogger(__name__)

class ArrowWriter(object):
    """Shuffles and writes Examples to Arrow files.
    """

    def __init__(self,
                 data_type: Optional[pa.DataType] = None,
                 schema: Optional[pa.Schema] = None,
                 path: Optional[str] = None,
                 stream: Optional[pa.NativeFile] = None,
                 writer_batch_size: Optional[int] = None,
                 disable_nullable: bool = True):
        if data_type is None and schema is None:
            raise ValueError("At least one of data_type and schema must be provided.")
        if path is None and stream is None:
            raise ValueError("At least one of path and stream must be provided.")

        if data_type is not None:
            self._type: pa.DataType = data_type
            self._schema: pa.Schema = pa.schema(field for field in self._type)
        else:
            self._schema: pa.Schema = schema
            self._type: pa.DataType = pa.struct(field for field in self._schema)

        if disable_nullable:
            self._schema = pa.schema(pa.field(field.name, field.type, nullable=False) for field in self._type)
            self._type = pa.struct(pa.field(field.name, field.type, nullable=False) for field in self._type)

        self._path = path
        if stream is None:
            self.stream = pa.OSFile(self._path, 'wb')
        else:
            self.stream = stream

        self.writer = pa.RecordBatchStreamWriter(self.stream, self._schema)
        self.writer_batch_size = writer_batch_size

        self._num_examples = 0
        self._num_bytes = 0
        self.current_rows = []

    def write_on_file(self):
        """ Write stored examples
        """
        pa_array = pa.array(self.current_rows, type=self._type)
        pa_batch = pa.RecordBatch.from_struct_array(pa_array)
        self._num_bytes += pa_array.nbytes
        self.writer.write_batch(pa_batch)
        self.current_rows = []

    def write(self, example: Dict[str, Any], writer_batch_size: Optional[int] = None):
        """ Add a given Example to the write-pool which is written to file.

        Args:
            example: the Example to add.
        """
        self.current_rows.append(example)
        self._num_examples += 1
        if writer_batch_size is None:
            writer_batch_size = self.writer_batch_size
        if writer_batch_size is not None and len(self.current_rows) >= writer_batch_size:
            self.write_on_file()

    def write_batch(self, batch_examples: Dict[str, List[Any]], writer_batch_size: Optional[int] = None):
        """ Write a batch of Example to file.

        Args:
            example: the Example to add.
        """
        pa_table: pa.Table = pa.Table.from_pydict(batch_examples, schema=self._schema)
        if writer_batch_size is None:
            writer_batch_size = self.writer_batch_size
        batches: List[pa.RecordBatch] = pa_table.to_batches(max_chunksize=writer_batch_size)
        self._num_bytes += sum(batch.nbytes for batch in batches)
        self._num_examples += pa_table.num_rows
        for batch in batches:
            self.writer.write_batch(batch)

    def finalize(self, close_stream=True):
        self.write_on_file()
        self.writer.close()
        if close_stream:
            self.stream.close()
        logger.info("Done writing %s examples in %s bytes %s.",
                    self._num_examples, self._num_bytes,
                    self._path if self._path else "")
        return self._num_examples, self._num_bytes


class BeamWriter(object):
    """Shuffles and writes Examples to Parquet files.
    """

    def __init__(self,
                 data_type: Optional[pa.DataType] = None,
                 schema: Optional[pa.Schema] = None,
                 path: Optional[str] = None,
                 stream: Optional[pa.NativeFile] = None,
                 writer_batch_size: Optional[int] = None):
        raise NotImplementedError



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Lazy imports for heavy dependencies."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib

from .utils import py_utils as utils


def _try_import(module_name):
    """Try importing a module, with an informative error message on failure."""
    try:
        mod = importlib.import_module(module_name)
        return mod
    except ImportError:
        err_msg = ("Failed importing {name}. This likely means that the dataset "
                             "requires additional dependencies that have to be "
                             "manually installed (usually with `pip install {name}`). See "
                             "setup.py extras_require.").format(name=module_name)
        utils.reraise(suffix=err_msg)


class LazyImporter(object):
    """Lazy importer for heavy dependencies.

    Some datasets require heavy dependencies for data generation. To allow for
    the default installation to remain lean, those heavy dependencies are
    lazily imported here.
    """

    @utils.classproperty
    @classmethod
    def apache_beam(cls):
        return _try_import("apache_beam")

    @utils.classproperty
    @classmethod
    def crepe(cls):
        return _try_import("crepe")

    @utils.classproperty
    @classmethod
    def cv2(cls):
        return _try_import("cv2")  # pylint: disable=unreachable

    @utils.classproperty
    @classmethod
    def h5py(cls):
        return _try_import("h5py")

    @utils.classproperty
    @classmethod
    def langdetect(cls):
        return _try_import("langdetect")

    @utils.classproperty
    @classmethod
    def librosa(cls):
        return _try_import("librosa")

    @utils.classproperty
    @classmethod
    def matplotlib(cls):
        _try_import("matplotlib.pyplot")
        return _try_import("matplotlib")

    @utils.classproperty
    @classmethod
    def mwparserfromhell(cls):
        return _try_import("mwparserfromhell")

    @utils.classproperty
    @classmethod
    def nltk(cls):
        return _try_import("nltk")

    @utils.classproperty
    @classmethod
    def pandas(cls):
        return _try_import("pandas")

    @utils.classproperty
    @classmethod
    def PIL_Image(cls):  # pylint: disable=invalid-name
        # TiffImagePlugin need to be activated explicitly on some systems
        # https://github.com/python-pillow/Pillow/blob/5.4.x/src/PIL/Image.py#L407
        _try_import("PIL.TiffImagePlugin")
        return _try_import("PIL.Image")

    @utils.classproperty
    @classmethod
    def pretty_midi(cls):
        return _try_import("pretty_midi")

    @utils.classproperty
    @classmethod
    def pydub(cls):
        return _try_import("pydub")

    @utils.classproperty
    @classmethod
    def scipy(cls):
        _try_import("scipy.io")
        _try_import("scipy.ndimage")
        return _try_import("scipy")

    @utils.classproperty
    @classmethod
    def skimage(cls):
        _try_import("skimage.color")
        _try_import("skimage.filters")
        _try_import("skimage.external.tifffile")
        return _try_import("skimage")

    @utils.classproperty
    @classmethod
    def tensorflow_io(cls):
        return _try_import("tensorflow_io")

    @utils.classproperty
    @classmethod
    def tldextract(cls):
        return _try_import("tldextract")

    @utils.classproperty
    @classmethod
    def os(cls):
        """For testing purposes only."""
        return _try_import("os")

    @utils.classproperty
    @classmethod
    def test_foo(cls):
        """For testing purposes only."""
        return _try_import("test_foo")


lazy_imports = LazyImporter  # pylint: disable=invalid-name



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""DatasetInfo records the information we know about a dataset.

This includes things that we know about the dataset statically, i.e.:
 - schema
 - description
 - canonical location
 - does it have validation and tests splits
 - size
 - etc.

This also includes the things that can and should be computed once we've
processed the dataset as well:
 - number of examples (in each split)
 - feature statistics (in each split)
 - etc.
"""

import abc
import collections
import json
import logging
import os
import shutil
import tempfile
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import numpy as np
import posixpath
from dataclasses_json import dataclass_json

from .splits import SplitDict, check_splits_equals

logger = logging.getLogger(__name__)

# Name of the file to output the DatasetInfo p rotobuf object.
DATASET_INFO_FILENAME = "dataset_info.json"
LICENSE_FILENAME = "LICENSE"

INFO_STR = """DatasetInfo(
        name='{name}',
        version={version},
        description='{description}',
        homepage='{homepage}',
        features={features},
        total_num_examples={total_num_examples},
        splits={splits},
        supervised_keys={supervised_keys},
        citation={citation},
        license={license},
)
"""


##### Replacing the p roto buff with a bunch of dataclasses for now
@dataclass
class SupervisedKeysData:
    input: str = ""
    output: str = ""


@dataclass
class RedistributionInfoData:
    license: str = ""

@dataclass
class DownloadChecksumsEntryData:
    key: str = ""
    value: str = ""

@dataclass_json
@dataclass
class DatasetInfoData:
    name: str = ""
    description: str = ""
    version: str = ""
    size_in_bytes: int = 0
    download_size: int = 0
    citation: Optional[str] = None
    homepage: Optional[str] = None
    license: Optional[str] = None
    splits: dict = field(default_factory=dict)
    supervised_keys: Optional[SupervisedKeysData] = None
    download_checksums: List[DownloadChecksumsEntryData] = field(default_factory=list)
    schema: Dict[str, Any] = field(default_factory=dict)


# TODO(nlp): Do we require to warn the user about the peak memory used while
# constructing the dataset?
class DatasetInfo(object):
    """Information about a dataset.

    `DatasetInfo` documents datasets, including its name, version, and features.
    See the constructor arguments and properties for a full list.

    Note: Not all fields are known on construction and may be updated later
    by `compute_dynamic_properties`. For example: the min and max values of a
    feature is typically updated during data generation (i.e. on calling
    builder.download_and_prepare()`).
    """

    def __init__(self,
                    builder,
                    description=None,
                    features=None,
                    supervised_keys=None,
                    homepage=None,
                    citation=None,
                    metadata=None,
                    license=None):
        """Constructs DatasetInfo.

        Args:
            builder: `DatasetBuilder`, dataset builder for this info.
            description: `str`, description of this dataset.
            features: `nlp.features.FeaturesDict`, Information on the feature dict
                of the `Dataset()` object from the `builder.as_dataset()`
                method.
            supervised_keys: `tuple` of `(input_key, target_key)`, Specifies the
                input feature and the label for supervised learning, if applicable for
                the dataset. The keys correspond to the feature names to select in
                `info.features`. When calling `nlp.DatasetBuilder.as_dataset()`
                with `as_supervised=True`, the `Dataset` object will yield
                the (input, target) defined here.
            homepage: `str`, optional, the homepage for this dataset.
            citation: `str`, optional, the citation to use for this dataset.
            metadata: dict, additonal object which will be
                stored/restored with the dataset. This allows for storing additional
                information with the dataset.
            license: `str`, optional,  will automatically be written to a
                LICENSE file stored with the dataset.
        """
        self._builder = builder

        self._info_data = DatasetInfoData(
                name=builder.name,
                description=description,
                version=str(builder._version),  # pylint: disable=protected-access
                citation=citation,
                homepage=homepage,
                splits=SplitDict(builder.name),
                license=license)

        if supervised_keys is not None:
            assert isinstance(supervised_keys, tuple)
            assert len(supervised_keys) == 2
            self._info_data.supervised_keys = SupervisedKeysData(input=supervised_keys[0],
                                                                 output=supervised_keys[1])

        self._features = features

        self._metadata = metadata

    @property
    def data(self):
        return self._info_data

    @property
    def name(self):
        return self.data.name

    @property
    def full_name(self):
        """Full canonical name: (<dataset_name>/<config_name>/<version>)."""
        names = [self._builder.name]
        if self._builder.builder_config:
            names.append(self._builder.builder_config.name)
        names.append(str(self.version))
        return posixpath.join(*names)

    @property
    def description(self):
        return self.data.description

    @property
    def version(self):
        return self._builder.version

    @property
    def homepage(self):
        return self.data.homepage

    @property
    def citation(self):
        return self.data.citation

    @property
    def data_dir(self):
        return self._builder.data_dir

    @property
    def dataset_size(self):
        """Generated dataset files size, in bytes."""
        # For old datasets, maybe empty.
        return sum(split.num_bytes for split in self.splits.values())

    @property
    def download_size(self):
        """Downloaded files size, in bytes."""
        # Fallback to deprecated `size_in_bytes` if `download_size` is empty.
        return self.data.download_size or self.data.size_in_bytes

    @download_size.setter
    def download_size(self, size):
        self.data.download_size = size

    @property
    def features(self):
        return self._features

    @property
    def metadata(self):
        return self._metadata

    @property
    def supervised_keys(self):
        if not self.data.supervised_keys is not None:
            return None
        supervised_keys = self.data.supervised_keys
        return (supervised_keys.input, supervised_keys.output)

    @property
    def license(self):
        return self.data.license

    @property
    def splits(self):
        return self.data.splits

    def update_splits_if_different(self, split_dict):
        """Overwrite the splits if they are different from the current ones.

        * If splits aren't already defined or different (ex: different number of
            shards), then the new split dict is used. This will trigger stats
            computation during download_and_prepare.
        * If splits are already defined in DatasetInfo and similar (same names and
            shards): keep the restored split which contains the statistics (restored
            from GCS or file)

        Args:
            split_dict: `nlp.SplitDict`, the new split
        """
        assert isinstance(split_dict, SplitDict)

        # If splits are already defined and identical, then we do not update
        if self.splits and check_splits_equals(
                self.splits, split_dict):
            return

        self.data.splits = split_dict.copy()

    def _dataset_info_path(self, dataset_info_dir):
        return os.path.join(dataset_info_dir, DATASET_INFO_FILENAME)

    def _license_path(self, dataset_info_dir):
        return os.path.join(dataset_info_dir, LICENSE_FILENAME)

    def write_to_directory(self, dataset_info_dir):
        """Write `DatasetInfo` as JSON to `dataset_info_dir`."""
        # Save the metadata from the features (vocabulary, labels,...)
        if self.features:
            self.features.save_metadata(dataset_info_dir)

        # Save any additional metadata
        if self.metadata is not None:
            self.metadata.save_metadata(dataset_info_dir)

        if self.data.license:
            with open(self._license_path(dataset_info_dir), "w") as f:
                f.write(self.data.license)

        with open(self._dataset_info_path(dataset_info_dir), "w") as f:
            f.write(self.data.to_json())

    def read_from_directory(self, dataset_info_dir):
        """Update DatasetInfo from the JSON file in `dataset_info_dir`.

        This function updates all the dynamically generated fields (num_examples,
        hash, time of creation,...) of the DatasetInfo.

        This will overwrite all previous metadata.

        Args:
            dataset_info_dir: `str` The directory containing the metadata file. This
                should be the root directory of a specific dataset version.
        """
        logger.info("Loading Dataset info from %s", dataset_info_dir)
        if not dataset_info_dir:
            raise ValueError(
                    "Calling read_from_directory with undefined dataset_info_dir.")

        dataset_info_path = self._dataset_info_path(dataset_info_dir)

        with open(dataset_info_path, "r") as f:
            # # Parse it back.
            dataset_info_json_str = f.read()
            self._info_data = DatasetInfoData.from_json(dataset_info_json_str)

        # Make sure we have a SplitDict back
        split_dict = self._info_data.splits
        self._info_data.splits = SplitDict.from_split_dict(split_dict)

        # Restore the feature metadata (vocabulary, labels names,...)
        if self.features:
            self.features.load_metadata(dataset_info_dir)

        if self.metadata is not None:
            self.metadata.load_metadata(dataset_info_dir)

        if self._builder._version != self.version:  # pylint: disable=protected-access
            raise AssertionError(
                    "The constructed DatasetInfo instance and the restored proto version "
                    "do not match. Builder version: {}. Proto version: {}".format(
                            self._builder._version, self.version))  # pylint: disable=protected-access

    def __repr__(self):
        splits_pprint = _indent("\n".join(["{"] + [
                "    '{}': {},".format(k, split.num_examples)
                for k, split in sorted(self.splits.items())
        ] + ["}"]))
        features_pprint = _indent(repr(self.features))
        citation_pprint = _indent('"""{}"""'.format(self.citation.strip()))
        return INFO_STR.format(
                name=self.name,
                version=self.version,
                description=self.description,
                total_num_examples=self.splits.total_num_examples,
                features=features_pprint,
                splits=splits_pprint,
                citation=citation_pprint,
                homepage=self.homepage,
                supervised_keys=self.supervised_keys,
                # Proto add a \n that we strip.
                license=str(self.license).strip())


def _indent(content):
    """Add indentation to all lines except the first."""
    lines = content.split("\n")
    return "\n".join([lines[0]] + ["    " + l for l in lines[1:]])



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Feature connector.

FeatureConnector is a way of abstracting what data is returned by the
datasets builders from how they are encoded/decoded from file.

# Use FeatureConnector in `GeneratorBasedBuilder`

1) In the _build_info() function, define the features as you would like them
to be returned by the pa.data.Dataset() object.

Ex:

    ```
    features=features.FeaturesDict({
            'input': features.Image(),
            'target': features.Text(encoder=SubWordEncoder()),
            'extra_data': {
                    'label_id': pa.int64,
                    'language': pa.string,
            }
    })
    ```

The pa.data.Dataset will return each examples as a dict:

    ```
    {
            'input': pa.Tensor(shape=(batch, height, width, channel), pa.uint8),
            'target': pa.Tensor(shape=(batch, sequence_length), pa.int64),
            'extra_data': {
                    'label_id': pa.Tensor(shape=(batch,), pa.int64),
                    'language': pa.Tensor(shape=(batch,), pa.string),
            }
    }
    ```

2) In the generator function, yield the examples to match what you have defined
in the spec. The values will automatically be encoded.

    ```
    yield {
            'input': np_image,
            'target': 'This is some text',
            'extra_data': {
                    'label_id': 43,
                    'language': 'en',
            }
    }
    ```

# Create your own FeatureConnector

To create your own feature connector, you need to inherit from FeatureConnector
and implement the abstract methods.

1. If your connector only contains one value, then the get_type,
     get_tensor_info, encode_example, and decode_example can directly process
     single value, without wrapping it in a dict.

2. If your connector is a container of multiple sub-connectors, the easiest
     way is to inherit from features.FeaturesDict and use the super() methods to
     automatically encode/decode the sub-connectors.

This file contains the following FeatureConnector:
 * FeatureConnector: The abstract base class defining the interface
 * FeaturesDict: Container of FeatureConnector
 * Tensor: Simple tensor value with static or dynamic shape

"""

import abc
import collections

import numpy as np

import pyarrow as pa


class TensorInfo(object):
    """Structure containing info on the `pa.Tensor` shape/dtype."""

    def __init__(self, shape, dtype, default_value=None, sequence_rank=None):
        """Constructor.

        Args:
            shape: `tuple[int]`, shape of the tensor
            dtype: Tensor dtype
            default_value: Used for retrocompatibility with previous files if a new
                field is added to provide a default value when reading the file.
            sequence_rank: `int`, Number of `nlp.features.Sequence` dimension.
        """
        self.shape = shape
        self.dtype = dtype
        self.default_value = default_value
        self.sequence_rank = sequence_rank or 0

    @classmethod
    def copy_from(cls, tensor_info):
        """Copy constructor."""
        return cls(
                shape=tensor_info.shape,
                dtype=tensor_info.dtype,
                default_value=tensor_info.default_value,
                sequence_rank=tensor_info.sequence_rank,
        )

    def __eq__(self, other):
        """Equality."""
        return (
                self.shape == other.shape and
                self.dtype == other.dtype and
                self.default_value == other.default_value
        )

    def __repr__(self):
        return '{}(shape={}, dtype={})'.format(
                type(self).__name__,
                self.shape,
                repr(self.dtype),
        )


class FeatureConnector(metaclass=abc.ABCMeta):
    """Abstract base class for feature types.

    This class provides an interface between the way the information is stored
    on disk, and the way it is presented to the user.

    Here is a diagram on how FeatureConnector methods fit into the data
    generation/reading:

    ```
    generator => encode_example() => tf_example => decode_example() => data dict
    ```

    The connector can either get raw or dictionary values as input, depending on
    the connector type.

    """

    @property
    def shape(self):
        """Return the shape (or dict of shape) of this FeatureConnector."""
        raise NotImplementedError

    @property
    def dtype(self):
        """Return the dtype (or dict of dtype) of this FeatureConnector."""
        return self.get_type()
        # return utils.map_nested(lambda t: t.dtype, self.get_type())

    # @abc.abstractmethod
    def get_type(self):
        """Return the shape/dtype of features after encoding (for the adapter).

        The `FileAdapter` then use those information to write data on disk.

        This function indicates how this feature is encoded on file internally.
        The DatasetBuilder are written on disk as pa.train.Example proto.

        Ex:

        ```
        return {
                'image': nlp.features.TensorInfo(shape=(None,), dtype=pa.uint8),
                'height': nlp.features.TensorInfo(shape=(), dtype=pa.int32),
                'width': nlp.features.TensorInfo(shape=(), dtype=pa.int32),
        }
        ```

        FeatureConnector which are not containers should return the feature proto
        directly:

        ```
        return nlp.features.TensorInfo(shape=(64, 64), pa.uint8)
        ```

        If not defined, the retuned values are automatically deduced from the
        `get_tensor_info` function.

        Returns:
            features: Either a dict of feature proto object, or a feature proto object

        """
        raise NotImplementedError

    @abc.abstractmethod
    def encode_example(self, example_data):
        """Encode the feature dict into tf-example compatible input.

        The input example_data can be anything that the user passed at data
        generation. For example:

        For features:

        ```
        features={
                'image': nlp.features.Image(),
                'custom_feature': nlp.features.CustomFeature(),
        }
        ```

        At data generation (in `_generate_examples`), if the user yields:

        ```
        yield {
                'image': 'path/to/img.png',
                'custom_feature': [123, 'str', lambda x: x+1]
        }
        ```

        Then:

         * `nlp.features.Image.encode_example` will get `'path/to/img.png'` as
             input
         * `nlp.features.CustomFeature.encode_example` will get `[123, 'str',
             lambda x: x+1] as input

        Args:
            example_data: Value or dictionary of values to convert into tf-example
                compatible data.

        Returns:
            tfexample_data: Data or dictionary of data to write as tf-example. Data
                can be a list or numpy array.
                Note that numpy arrays are flattened so it's the feature connector
                responsibility to reshape them in `decode_example()`.
                Note that pa.train.Example only supports int64, float32 and string so
                the data returned here should be integer, float or string. User type
                can be restored in `decode_example()`.
        """
        raise NotImplementedError

    def decode_example(self, tfexample_data):
        """Decode the feature dict to TF compatible input.

        Note: If eager is not enabled, this function will be executed as a
        tensorflow graph (in `pa.data.Dataset.map(features.decode_example)`).

        Args:
            tfexample_data: Data or dictionary of data, as read by the tf-example
                reader. It correspond to the `pa.Tensor()` (or dict of `pa.Tensor()`)
                extracted from the `pa.train.Example`, matching the info defined in
                `get_type()`.

        Returns:
            tensor_data: Tensor or dictionary of tensor, output of the pa.data.Dataset
                object
        """
        return tfexample_data

    def decode_batch_example(self, tfexample_data):
        """Decode multiple features batched in a single pa.Tensor.

        This function is used to decode features wrapped in
        `nlp.features.Sequence()`.
        By default, this function apply `decode_example` on each individual
        elements using `pa.map_fn`. However, for optimization, features can
        overwrite this method to apply a custom batch decoding.

        Args:
            tfexample_data: Same `pa.Tensor` inputs as `decode_example`, but with
                and additional first dimension for the sequence length.

        Returns:
            tensor_data: Tensor or dictionary of tensor, output of the pa.data.Dataset
                object
        """
        # Note: This all works fine in Eager mode (without pa.function) because
        # pa.data pipelines are always executed in Graph mode.

        # Apply the decoding to each of the individual distributed features.
        return tf.map_fn(
                self.decode_example,
                tfexample_data,
                dtype=self.dtype,
                parallel_iterations=10,
                back_prop=False,
                name='sequence_decode',
        )


    def decode_ragged_example(self, tfexample_data):
        """Decode nested features from a pa.RaggedTensor.

        This function is used to decode features wrapped in nested
        `nlp.features.Sequence()`.
        By default, this function apply `decode_batch_example` on the flat values
        of the ragged tensor. For optimization, features can
        overwrite this method to apply a custom batch decoding.

        Args:
            tfexample_data: `pa.RaggedTensor` inputs containing the nested encoded
                examples.

        Returns:
            tensor_data: The decoded `pa.RaggedTensor` or dictionary of tensor,
                output of the pa.data.Dataset object
        """
        return tf.ragged.map_flat_values(self.decode_batch_example, tfexample_data)

    def _flatten(self, x):
        """Flatten the input dict into a list of values.

        For instance, the following feature:
        ```
        feature = FeatureDict({
                'a': w,
                'b': x,
                'c': {
                        'd': y,
                        'e': z,
                },
        })
        ```

        Applied to the following `dict`:
        ```
        feature._flatten({
                'b': X,
                'c': {
                        'd': Y,
                },
        })
        ```

        Will produce the following flattened output:
        ```
        [
                None,
                X,
                Y,
                None,
        ]
        ```

        Args:
            x: A nested `dict` like structure matching the structure of the
            `FeatureConnector`. Note that some elements may be missing.

        Returns:
            `list`: The flattened list of element of `x`. Order is guaranteed to be
            deterministic. Missing elements will be filled with `None`.
        """
        return [x]

    def _nest(self, list_x):
        """Pack the list into a nested dict.

        This is the reverse function of flatten.

        For instance, the following feature:
        ```
        feature = FeatureDict({
                'a': w,
                'b': x,
                'c': {
                        'd': y,
                        'e': z,
                },
        })
        ```

        Applied to the following `dict`:
        ```
        feature._nest([
                None,
                X,
                Y,
                None,
        ])
        ```

        Will produce the following flattened output:
        ```
        {
                'a': None,
                'b': X,
                'c': {
                        'd': Y,
                        'e': None,
                },
        }
        ```

        Args:
            list_x: List of values matching the flattened `FeatureConnector`
                structure. Missing values should be filled with None.

        Returns:
            nested_x: nested `dict` matching the flattened `FeatureConnector`
                structure.
        """
        assert len(list_x) == 1
        return list_x[0]

    def _additional_repr_info(self):
        """Override to return additional info to go into __repr__."""
        return {}

    def __repr__(self):
        """Display the feature dictionary."""
        tensor_info = self.get_type()
        return str(tensor_info)
        # if not isinstance(tensor_info, TensorInfo):
        #     return '{}({})'.format(type(self).__name__, tensor_info)

        # # Ensure ordering of keys by adding them one-by-one
        # repr_info = collections.OrderedDict()
        # repr_info['shape'] = tensor_info.shape
        # repr_info['dtype'] = repr(tensor_info.dtype)
        # additional_info = self._additional_repr_info()
        # for k, v in additional_info.items():
        #     repr_info[k] = v

        # info_str = ', '.join(['%s=%s' % (k, v) for k, v in repr_info.items()])
        # return '{}({})'.format(
        #         type(self).__name__,
        #         info_str,
        # )

    def save_metadata(self, data_dir, feature_name):
        """Save the feature metadata on disk.

        This function is called after the data has been generated (by
        `_download_and_prepare`) to save the feature connector info with the
        generated dataset.

        Some dataset/features dynamically compute info during
        `_download_and_prepare`. For instance:

         * Labels are loaded from the downloaded data
         * Vocabulary is created from the downloaded data
         * ImageLabelFolder compute the image dtypes/shape from the manual_dir

        After the info have been added to the feature, this function allow to
        save those additional info to be restored the next time the data is loaded.

        By default, this function do not save anything, but sub-classes can
        overwrite the function.

        Args:
            data_dir: `str`, path to the dataset folder to which save the info (ex:
                `~/datasets/cifar10/1.2.0/`)
            feature_name: `str`, the name of the feature (from the FeaturesDict key)
        """
        pass

    def load_metadata(self, data_dir, feature_name):
        """Restore the feature metadata from disk.

        If a dataset is re-loaded and generated files exists on disk, this function
        will restore the feature metadata from the saved file.

        Args:
            data_dir: `str`, path to the dataset folder to which save the info (ex:
                `~/datasets/cifar10/1.2.0/`)
            feature_name: `str`, the name of the feature (from the FeaturesDict key)
        """
        pass


class Tensor(FeatureConnector):
    """`FeatureConnector` for generic data of arbitrary shape and type.
        We are deprecating 'shape'
    """

    def __init__(self, shape, dtype):
        """Construct a Tensor feature."""
        assert len(shape) < 2, "Tensor can only take 0 or 1 dimensional shapes ."
        self._shape = tuple(shape)
        if len(shape) == 1:
            self._dtype = pa.list_(dtype())
        else:
            self._dtype = dtype()

    def get_type(self) -> pa.DataType:
        """See base class for details."""
        return self._dtype

    def decode_batch_example(self, example_data):
        """See base class for details."""
        # Overwrite the `pa.map_fn`, decoding is a no-op
        return self.decode_example(example_data)

    def decode_ragged_example(self, example_data):
        """See base class for details."""
        # Overwrite the `pa.map_fn`, decoding is a no-op
        return self.decode_example(example_data)

    def encode_example(self, example_data):
        """See base class for details."""
        # np_dtype = np.dtype(self.dtype.to_pandas_dtype())
        # if not isinstance(example_data, np.ndarray):
        #     example_data = np.array(example_data, dtype=np_dtype)
        # # Ensure the shape and dtype match
        # if example_data.dtype != np_dtype:
        #     raise ValueError('Dtype {} do not match {}'.format(
        #             example_data.dtype, np_dtype))
        # utils.assert_shape_match(example_data.shape, self._shape)
        return example_data
